/* This file is part of Jumper
   Copyright (C) 2013 Sergey Poznyakoff
  
   Jumper is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.
  
   Jumper is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with Jumper.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "jumper.h"
#include <sys/wait.h>

int snaplen = 96;
int promisc_option = 1;

struct callback {
	pcap_handler    function;
	int             type;
};


static struct callback callbacks[] = {
	{ cons_ether_packet, DLT_EN10MB },
	{ cons_ether_packet, DLT_IEEE802 },
	{ cons_raw_packet, DLT_RAW },
	{ cons_ppp_packet, DLT_PPP },
	{ NULL, DLT_FDDI },
	{ NULL, DLT_SLIP },
	{ NULL, DLT_NULL },
	{ NULL, 0 },
};

pcap_handler
lookup_pcap_callback(int type)
{
	struct callback *callback;

	for (callback = callbacks; callback->function; callback++)
		if (type == callback->type)
			return callback->function;
	diag(LOG_ERR, "unknown data link type 0x%x", type);
	exit(EX_SOFTWARE);
}

struct interface {
	struct interface *next;
	char *name;
	pcap_t *pcap;
	pcap_handler hfun;
	char *expr;
};

static struct interface *if_head, *if_tail;

struct interface *
interface_lookup(const char *name)
{
	struct interface *ip;
	
	for (ip = if_head; ip; ip = ip->next) {
		if (strcmp(ip->name, name) == 0)
			return ip;
	}

	ip = emalloc(sizeof(*ip));
	ip->next = NULL;
	ip->pcap = NULL;
	ip->name = estrdup(name);
	ip->expr = NULL;

	if (if_tail)
		if_tail->next = ip;
	else
		if_head = ip;
	if_tail = ip;
	return ip;
}

void
interface_init()
{
	struct interface *iface;
	pcap_t *pd;
	char errbuf[PCAP_ERRBUF_SIZE];
	
	if (!if_head) {
		diag(LOG_CRIT, "no listeners configured");
		exit(EX_CONFIG);
	}

	for (iface = if_head; iface; iface = iface->next) {
		pd = pcap_open_live(iface->name, snaplen, !promisc_option,
				    1000, errbuf);
		if (!pd) {
			diag(LOG_CRIT, "pcap_open_live: %s", errbuf);
			exit(EX_OSERR);
		}
		iface->pcap = pd;
		iface->hfun = lookup_pcap_callback(pcap_datalink(pd));
	}
}


listener_t *l_head, *l_tail;

void
listener_fixup(listener_t *lp)
{
 	lp->prog = estrdup(lp->prog);
	lp->locus.beg.file = estrdup(lp->locus.beg.file);
	lp->locus.end.file = estrdup(lp->locus.end.file);
}

void
listener_add(listener_t *lp)
{
	if (l_tail)
		l_tail->next = lp;
	else
		l_head = lp;
	l_tail = lp;
}

static const char *
ipv4_to_string(uint32_t ip)
{
	static char buf[16];
	
	sprintf(buf, "%u.%u.%u.%u",
		(ip >> 24) & 0xff,
		(ip >> 16) & 0xff,
		(ip >> 8) & 0xff,
		ip & 0xff);

	return buf;
}

#define addstr(s) grecs_line_acc_grow(s, strlen(s))

void
listener_init()
{
	struct interface *ifp;
	listener_t *lp;
	ipv4_match_list_t *mp;
	int delim;
	struct bpf_program bpfcode;
	
	grecs_line_acc_create();
	for (lp = l_head; lp; lp = lp->next) {
		grecs_line_begin();
		if (lp->iface->expr) {
			addstr(lp->iface->expr);
			addstr(" or ");
		}
		addstr(" ( ");
		delim = 0;
		if (lp->match_source) {
			addstr(" ( ");
			for (mp = lp->match_source; mp; mp = mp->next) {
				if (delim)
					addstr(" or");
				addstr(" src net ");
				addstr(ipv4_to_string(mp->cidr.addr));
				addstr(" mask ");
				addstr(ipv4_to_string(mp->cidr.netmask));
				delim = 1;
			}
			addstr(" ) and ");
		}
		delim = 0;
		addstr(" ( ");
		for (mp = lp->match_dest; mp; mp = mp->next) {
			if (delim)
				addstr(" or");
			addstr(" dst net ");
			addstr(ipv4_to_string(mp->cidr.addr));
			addstr(" mask ");
			addstr(ipv4_to_string(mp->cidr.netmask));
			delim = 1;
		}
		addstr(" ) ");
		addstr(" ) ");
		lp->iface->expr = grecs_line_finish();
	}
	
	for (ifp = if_head; ifp; ifp = ifp->next) {
		debug(2, ("iface %s, expr %s", ifp->name, ifp->expr));
		
		if (pcap_compile(ifp->pcap, &bpfcode, ifp->expr, 1, 0) < 0) {
			diag(LOG_ERR, "pcap_compile(%s): %s",
			     ifp->expr, pcap_geterr(ifp->pcap));
			abort();
		}

		if (pcap_setfilter(ifp->pcap, &bpfcode) < 0) {
			diag(LOG_ERR, "pcap_setfilter: %s",
			     pcap_geterr(ifp->pcap));
			exit(EX_OSERR);
		}
	}
	grecs_line_acc_free();
}


void
listener_kve_init(char **kve, int kvn, listener_t *lp, ...)
{
	int i = 0;
	static char linebuf[64];
	va_list ap;
	char *p;
	
	if (kvn < KVE_MINSIZE) {
		diag(LOG_CRIT, "%s:%d: INTERNAL ERROR", __FILE__, __LINE__);
		abort();
	}
	
	kve[i++] = "program";
	kve[i++] = (char*)program_name;

	kve[i++] = "file";
	kve[i++] = lp->locus.beg.file;

	kve[i++] = "line";
	snprintf(linebuf, sizeof(linebuf), "%u", lp->locus.beg.line);
	kve[i++] = linebuf;

	va_start(ap, lp);
	while ((p = va_arg(ap,char*))) {
		if (i == kvn) {
			diag(LOG_CRIT, "%s:%d: INTERNAL ERROR",
			     __FILE__, __LINE__);
			abort();
		}
		kve[i++] = p;
	}
	va_end(ap);
	kve[i] = NULL;
}

static char *event_str[] = {
	"startup",
	"cleanup",
	"exit",
	"signal"
};

static event_t *
event_match_type(action_t *act, enum event_type type)
{
	event_t *evt;

	for (evt = act->evt; evt; evt = evt->next)
		if (evt->type == type)
			break;
	return evt;
}

static event_t *
event_match(action_t *act, enum event_type type, int code)
{
	event_t *evt;

	for (evt = act->evt; evt; evt = evt->next)
		if (evt->type == type && evt->code == code)
			break;
	return evt;
}

int
listener_run_action(int a)
{
	listener_t *lp;
	char *kve[KVE_MINSIZE+2];
	int err;
	
	for (lp = l_head; lp; lp = lp->next) {
		action_t *act;
		
		listener_kve_init(kve, sizeof(kve)/sizeof(kve[0]),
				  lp,
				  "event", event_str[a],
				  NULL);

		for (act = lp->act_head; act; act = act->next) {
			if (event_match_type(act, a) &&
			    progman_start(type_action, lp,
					  &act->locus,
					  act->options,
					  act->timeout,
					  act->command,
					  act->env, kve))
				err++;
		}
	}
	return err;
}


void
listener_run()
{
	struct interface *ifp;
	fd_set rdset;
	int maxfd = -1;
	int fd;
	int rc;
	struct timeval tv;
	
	listener_proc_report();
	
	FD_ZERO(&rdset);
	for (ifp = if_head; ifp; ifp = ifp->next) {
		fd = pcap_fileno(ifp->pcap);
		if (fd > maxfd)
			maxfd = fd;
		FD_SET(fd, &rdset);
	}

	rc = select(maxfd + 1, &rdset, NULL, NULL, progman_timeout(&tv));
	if (rc < 0) {
		if (errno == EINTR || errno == EAGAIN)
			return;
		diag(LOG_ERR, "select: %s", strerror(errno));
		exit(EX_OSERR);
	} else if (rc == 0) {
		progman_expire();
		return;
	}

	for (ifp = if_head; ifp; ifp = ifp->next) {
		fd = pcap_fileno(ifp->pcap);
		if (FD_ISSET(fd, &rdset)) {
			struct pcap_pkthdr hdr;
			const u_char *data = pcap_next(ifp->pcap, &hdr);
			if (!data)
				continue;
			ifp->hfun(NULL, &hdr, data);
		}
	}
}

int
ipv4_match_p(ipv4_match_list_t *mp, uint32_t ip)
{
	for (; mp; mp = mp->next)
		if (mp->cidr.addr == (mp->cidr.netmask & ip))
			return 1;
	return 0;
}

listener_t *
listener_locate(uint32_t src, uint32_t dst)
{
	listener_t *lp;

	for (lp = l_head; lp; lp = lp->next) {
		if ((!lp->match_source ||
		     ipv4_match_p(lp->match_source, src)) &&
		    ipv4_match_p(lp->match_dest, dst))
			break;
	}
	return lp;
}

listener_t *
listener_lookup_pid(pid_t pid)
{
	listener_t *lp;
	for (lp = l_head; lp; lp = lp->next)
		if (lp->pid == pid)
			break;
	return lp;
}

void
listener_print_status(listener_t *lp)
{
	if (WIFEXITED(lp->progstat)) {
		if (WEXITSTATUS(lp->progstat) == 0)
			debug(1, ("%s:%d: %s (%lu) exited successfully",
				  lp->locus.beg.file, lp->locus.beg.line,
				  lp->prog, (unsigned long) lp->pid));
		else
			diag(LOG_NOTICE,
			     "%s:%d: %s (%lu) exited with status %d",
			     lp->locus.beg.file, lp->locus.beg.line,
			     lp->prog, (unsigned long) lp->pid,
			     WEXITSTATUS(lp->progstat));
	} else if (WIFSIGNALED(lp->progstat))
		diag(LOG_NOTICE,
		     "%s:%d: %s (%lu) terminated on signal %d",
		     lp->locus.beg.file, lp->locus.beg.line,
		     lp->prog, (unsigned long) lp->pid,
		     WTERMSIG(lp->progstat));
	else if (WIFSTOPPED(lp->progstat))
		diag(LOG_NOTICE,
		     "%s:%d: %s (%lu) stopped",
		     lp->locus.beg.file, lp->locus.beg.line,
		     lp->prog, (unsigned long) lp->pid);

#ifdef WCOREDUMP
	if (WCOREDUMP(lp->progstat))
		diag(LOG_NOTICE,
		     "%s:%d: %s (%lu) dumped core",
		     lp->locus.beg.file, lp->locus.beg.line,
		     lp->prog, (unsigned long) lp->pid);
#endif

	lp->status = stat_down;
} 

static void
runaction(listener_t *lp, action_t *act, event_t *evt)
{
	char *kve[KVE_MINSIZE + 6];
	char codebuf[64];
	char statbuf[64];

	snprintf(statbuf, sizeof(statbuf), "%u", lp->progstat);
	snprintf(codebuf, sizeof(codebuf), "%u", evt->code);
	listener_kve_init(kve, sizeof(kve)/sizeof(kve[0]),
			  lp,
			  "event", event_str[evt->type],
			  "status", statbuf,
			  "exitcode", codebuf,
			  NULL);

	if (progman_start(type_action, lp, &act->locus, act->options,
			  act->timeout, act->command, act->env, kve))
		return;

	lp->act_count++;
}

void
onexit_reaction(listener_t *lp)
{
	action_t *act;
	int code;
	enum event_type type;
	event_t *evt;
	
	lp->status = stat_onexit;
	if (WIFEXITED(lp->progstat)) {
		type = event_exit;
		code = WEXITSTATUS(lp->progstat);
	} else if (WIFSIGNALED(lp->progstat)) {
		type = event_signal;
		code = WTERMSIG(lp->progstat);
	}

	for (act = lp->act_head; act; act = act->next) {
		if (evt = event_match(act, type, code))
			runaction(lp, act, evt);
	}
	if (lp->act_count == 0)
		lp->status = stat_down;
}

void
listener_proc_report(void)
{
	listener_t *lp;
	
	if (proc_stop == 0)
		return;
	proc_stop = 0;

	for (lp = l_head; lp; lp = lp->next) {
		if (lp->status == stat_term) {
			listener_print_status(lp);
			onexit_reaction(lp);
		}
	}
}

void
listener_start(listener_t *lp, char **kve)
{
	progman_start(type_listener, lp, &lp->locus, lp->options, 0, 
		      lp->prog, lp->env, kve);
}

